#include <cmath>
#include <cstdlib> //for random numbers: rand();
#include <SDL.h>
#include "QuickCG.h"
#include "Q3DMath.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

using namespace std;
struct vector3d{
       float x, y, z;
       };
struct ray{
       vector3d o, d;
       };
struct Material{
       float dif, refl, refr;
       ColorRGB col;
       };
struct sphere{
       Material mat;
       vector3d c;
       float r;
       };
       
struct plane{
       Material mat;
       vector3d a, b, c, n;
       };
struct tri{
       Material mat;
       vector3d a, b, c, n;
       };
vector3d Vector3d(float x, float y, float z){
         vector3d ret;
         ret.x = x;
         ret.y = y;
         ret.z = z;
         return ret;
         }
vector3d normalize(vector3d v){
         float len = sqrt(v.x*v.x+v.y*v.y+v.z*v.z);
         vector3d n;
         n.x = v.x/len;
         n.y = v.y/len;
         n.z = v.z/len;
         return n;
         }
float dot(vector3d v, vector3d w){
         return v.x*w.x + v.y*w.y +v.z*w.z;
         }
 
vector3d cross(vector3d a, vector3d b){
         vector3d c = Vector3d(a.y*b.z-a.z*b.y, a.z*b.x-a.x*b.z, a.x*b.y-a.y*b.x);
         return normalize(c);
         }
double PlaneIntersect(plane p, ray r)
{
    float D = sqrt(p.c.x*p.c.x+p.c.y*p.c.y+p.c.z*p.c.z);
    float numer = dot(p.n, r.o) + D;
    float denom = dot(p.n, r.d);
    return -(numer / denom);
}
double SphereIntersect(sphere s, ray r){
       float b = 2*r.d.x*(r.o.x-s.c.x)+2*r.d.y*(r.o.y-s.c.y)+2*r.d.z*(r.o.z-s.c.z);
       float c = (r.o.x-s.c.x)*(r.o.x-s.c.x)+(r.o.y-s.c.y)*(r.o.y-s.c.y)+(r.o.z-s.c.z)*(r.o.z-s.c.z)-s.r*s.r;
       float det = b*b-4*c;
       if (det < 0){
               return -1;
               }
       if (det == 0){
               return 0;
               }
       if (det > 0){
          float t1, t2;
          t1 = (-b-sqrt(b*b-4*c))/2;
          t2 = (-b+sqrt(b*b-4*c))/2;
          if (t1 < 0){
                 return t2;
                 }
          else if (t1 > 0){
               return t1;
               }
               }
               }
               
vector3d GetPP(float t, ray r){//R(t) = R0 + tRd
         return Vector3d(r.o.x+r.d.x*t, r.o.y+r.d.y*t, r.o.z+r.d.z*t);
         }
vector3d GetNormal(vector3d p, sphere s){
         return normalize(Vector3d(p.x-s.c.x, p.y-s.c.y, p.z-s.c.z));
         }
ColorRGB fix(ColorRGB col){
         if (col.r > 255){
                   col.r = 255;
                   }
         if(col.r < 0){
                  col.r = 0;
                  }
         if (col.g > 255){
                   col.g = 255;
                   }
         if (col.g < 0){
                   col.g = 0;
                   }
         if (col.b > 255){
                   col.b = 255;
                   }
         if (col.b < 0){
                   col.b = 0;
                   }
         return col;
         }
float TriIntersect(tri t, ray r){
      float numer = dot(Vector3d(r.o.x - t.a.x, r.o.y-t.a.y, r.o.z-t.a.z), t.n);
      float denom = dot(r.d, t.n);
      float time;
      time = -(numer/denom);
      vector3d pp = GetPP(time, r);
      vector3d b = Vector3d(t.b.x-t.a.x, t.b.y-t.a.y, t.b.z-t.a.z);
      vector3d c = Vector3d(t.c.x-t.a.x, t.c.y-t.a.y, t.c.z-t.a.y);
      vector3d p = Vector3d(pp.x-t.a.x, pp.y-t.a.y, pp.z-t.a.z);
      float u = (p.y*c.x-p.x*c.y)/(b.y*c.x-b.x*c.y);
      float v = (p.y*b.x-p.x*b.y)/(c.y*b.x-c.x*b.y);
      if (u >= 0 && v >= 0 && u+v <= 1){
            return time;
            }
      else{
           return -1;
           }
           }
float GetDS(vector3d light, vector3d norm, vector3d pi){
      vector3d l = normalize(Vector3d(light.x-pi.x, light.y-pi.y, light.z-pi.z));
      return dot(l, norm);
      }
float GetShadeS(vector3d light, sphere s, vector3d o){
      vector3d temp = Vector3d(light.x-o.x, light.y-o.y, light.z-o.z);
      float len = sqrt(temp.x*temp.x+temp.y*temp.y+temp.z*temp.z);
      ray shad;
      shad.o = o;
      shad.d = normalize(temp);
      float st = SphereIntersect(s, shad);
      if (st > 0 && st < len){
             return 0;
             }
      else {
           return 1;
           }
           }
ray GetReflected(vector3d o, ray i, vector3d N){
    ray out;
    out.o = o;
    out.d.x = i.d.x - 2* dot(i.d, N) * N.x; 
    out.d.y = i.d.y - 2* dot(i.d, N) * N.y;
    out.d.z = i.d.z - 2* dot(i.d, N) * N.z;
    out.d = normalize(out.d);
    return out;
}  
int main(int argc, char *argv[])
{
    //////////////////////samples
    int samples = 16;
    float sqrsamp = sqrt(samples);
    vector3d shadtar;
    int shadSamps = 14;
    int ShadRays = shadSamps*shadSamps;
    //////////////////////main ray
    ray r;
    r.o = Vector3d(-5,0, 3);
    ////////////////////////ground plane
    plane ground;
    ground.a = Vector3d(1,0,0);
    ground.b = Vector3d(0,1,0);
    ground.c = Vector3d(0,0,0);
    ground.n = cross(ground.a, ground.b);
    ground.mat.dif = 1;
    ground.mat.col = ColorRGB(200,200,200);
    /////////////////sphere 1
    sphere s;
    s.c = Vector3d(5,-3,1);
    s.r = 1;
    s.mat.dif = 1;
    s.mat.refl = 1;
    s.mat.col = ColorRGB(100,100,255);
    sphere s2;
    s2.c = Vector3d(7, 3,3);
    s2.r = 3;
    s2.mat.dif = 1;
    s2.mat.col = ColorRGB(100,100,255);
    
    ////////////////2 triangles for area light
    tri tri1;
    tri1.a = Vector3d(12, -3, 10);
    tri1.b = Vector3d(12, 3, 10);
    tri1.c = Vector3d(6, -3, 10);
    tri1.n = Vector3d(0,0,-1);
    tri1.mat.dif = 1;
    tri1.mat.col = ColorRGB(255,255,255);
    tri tri2;
    tri2.a = Vector3d(12,3,10);
    tri2.b = Vector3d(6,3,10);
    tri2.c = Vector3d(6,-3,10);
    tri2.n = Vector3d(0,0,-1); 
    tri2.mat.dif = 1;
    tri2.mat.col = ColorRGB(255,255,255);
    vector3d alight = Vector3d(9,0,10);
    shadtar = Vector3d(6,3,10);
    ////////////////t values
    float t = 10000000, t2 = 0;
    vector3d p;
    vector3d sn;
    ////////////////main color
    ColorRGB pixcol;
    ColorRGB aacol[5];
    float ds;
    float tempshade;
    float shade=1;
    float St;
    float jitter, jitter2;
    int posneg, posneg2;
    vector3d jittar;
    float shadAVG[ShadRays], shadAVG2[ShadRays];
    int shadct = 0;
    float tempshade2 = 0;
          float shade2 = 0;
          ray refl;
          float rt, rt2;
          float SS;
    /////////////////screen
    vector3d scr = Vector3d(0,-4, 6);
    screen(800, 600, 0, "IceRay v0.0.1");
        vector3d backup;
    /////////////////main loop
    srand(time(NULL));
    for(int y = 0; y < 600; y++){
    scr.y=-4;
    
    for(int x = 0; x < 800; x++)
    {
       for (int sa=0; sa < 5; sa++){
       aacol[sa] = ColorRGB(0,0,0);
       shadct = 0;
       shadtar = Vector3d(6,3,10);
       shade = 1;
       tempshade2 = 0;
                 shade2 = 0;
       tempshade = 1;
       jittar = Vector3d(0,0,0);
       posneg = 0;
       jitter = 0;
       t = 10000000;
       
       r.d = (Vector3d(scr.x-r.o.x, scr.y-r.o.y, scr.z-r.o.z)); 
       r.d = normalize(r.d);
       backup = r.d;
       if (sa ==1){
       r.d = (Vector3d(scr.x-r.o.x, scr.y-0.005-r.o.y, scr.z+0.005-r.o.z)); 
       r.d = normalize(r.d);
       }
       if (sa ==2){
       r.d = (Vector3d(scr.x-r.o.x, scr.y-0.005-r.o.y, scr.z-0.005-r.o.z)); 
       r.d = normalize(r.d);
       }
       if (sa ==3){
       r.d = (Vector3d(scr.x-r.o.x, scr.y+0.005+r.o.y, scr.z+0.005-r.o.z)); 
       r.d = normalize(r.d);
       }
       if (sa ==4){
       r.d = (Vector3d(scr.x-r.o.x, scr.y+0.005-r.o.y, scr.z-0.005-r.o.z)); 
       r.d = normalize(r.d);
       }
       for (int reset = 0; reset < ShadRays; reset++){
                     shadAVG[reset]=1;
                     }
       t2 = SphereIntersect(s, r);
       if (t2 > 0&&t2 < t){
              t = t2;
              p = GetPP(t, r);
              sn = GetNormal(p, s);
              ds = GetDS(alight, sn, p);
              refl = GetReflected(p, r, sn);
              if (ds > 0){
              aacol[sa] = ColorRGB(s.mat.col.r*ds, s.mat.col.g*ds, s.mat.col.b*ds);
              }
              else{
                   aacol[sa] = ColorRGB(0,0,0);
                   }
              if (s.mat.refl > 0){
                               aacol[sa] = ColorRGB(0,0,0);
                               }
              if (s.mat.refl > 0){
                            rt=1000000;
                            rt2 = SphereIntersect(s2, refl);
                             if (rt2 < rt && rt2 > 0){
                                     rt = rt2;
              p = GetPP(rt, refl);
              sn = GetNormal(p, s2);
              ds = GetDS(alight, sn, p);
              if (ds > 0){
              aacol[sa] = ColorRGB(s2.mat.col.r*ds, s2.mat.col.g*ds, s2.mat.col.b*ds);
              }
              else {
                   aacol[sa] = ColorRGB(0,0,0);
                   }
              }
              rt2 = PlaneIntersect(ground, refl);
              if (rt2 < rt && rt2 > 0){
                      rt = rt2;
              shadtar = Vector3d(6,3,9);
              rt = rt2;
              p = GetPP(rt, refl);
              ds = GetDS(alight, ground.n, p);
              if (ds > 0){
              aacol[sa] = fix(ColorRGB(ground.mat.col.r*ds, ground.mat.col.g*ds, ground.mat.col.b*ds));
              }
              else if (ds < 0){
                   aacol[sa] = ColorRGB(0,0,0);
                   }
                     for (int sx = 0;sx < shadSamps;sx++){
                         shadtar.y = 0;
                                    for (int sy = 0; sy < shadSamps; sy++){
                                    jitter = ((rand()%12501)/1000);
                                    posneg = rand()%2;
                                    jitter2 = ((rand()%12501)/1000);
                                    posneg2 = rand()%2;
                                    if(posneg==1){
                                                  jitter= -jitter;
                                                  }
                                                  if(posneg2==1){
                                                  jitter2= -jitter2;
                                                  }
                                    jittar = Vector3d(shadtar.x+jitter, shadtar.y+jitter2, 10);
                                    tempshade = GetShadeS(jittar, s, p);
                                    shadAVG[shadct]=tempshade;
                                    shadtar.y = shadtar.y - 6/shadSamps;
                                    shadct++;
                                    }
                         shadtar.x = shadtar.x + 6/shadSamps;
                         }
                         shadtar = Vector3d(6,3,10);
                         shadct=0;
                         for (int sx = 0;sx < shadSamps;sx++){
                         shadtar.y = 0;
                                    for (int sy = 0; sy < shadSamps; sy++){
                                    jitter = ((rand()%12501)/1000);
                                    posneg = rand()%2;
                                    jitter2 = ((rand()%12501)/1000);
                                    posneg2 = rand()%2;
                                    if(posneg==1){
                                                  jitter= -jitter;
                                                  }
                                                  if(posneg2==1){
                                                  jitter2= -jitter2;
                                                  }
                                    jittar = Vector3d(shadtar.x+jitter, shadtar.y+jitter2, 10);
                                    tempshade = GetShadeS(jittar, s2, p);
                                    shadAVG2[shadct]=tempshade;
                                    shadtar.y = shadtar.y - 6/shadSamps;
                                    shadct++;
                                    }
                         shadtar.x = shadtar.x + 6/shadSamps;
                         }
                 tempshade = 0;
                 tempshade2 = 0;
                 shade2 = 0;
                 for (int avg= 0; avg < ShadRays; avg++){
                     tempshade += shadAVG[avg];
                     tempshade2 += shadAVG2[avg];
                     }
                 shade = tempshade/ShadRays;
                 shade2 = tempshade2/ShadRays;
                 aacol[sa]=ColorRGB(aacol[sa].r*shade*shade2, aacol[sa].g*shade*shade2, aacol[sa].b*shade*shade2);
              }
              rt2 = TriIntersect(tri1, refl);
              if (rt2 < rt && rt2 > 0){
              rt = rt2;
              aacol[sa] = tri1.mat.col;
              }
              rt2 = TriIntersect(tri2, refl);
              if (rt2 < rt && rt2 > 0){
              rt = rt2;
              aacol[sa] = tri2.mat.col;
              }
              }
              }
       t2 = PlaneIntersect(ground, r);
       if (t2 > 0&&t2 < t){
              shadtar = Vector3d(6,3,9);
              t = t2;
              p = GetPP(t2, r);
              ds = GetDS(alight, ground.n, p);
              if (ds > 0){
              aacol[sa] = fix(ColorRGB(ground.mat.col.r*ds, ground.mat.col.g*ds, ground.mat.col.b*ds));
              }
              else if (ds < 0){
                   aacol[sa] = ColorRGB(0,0,0);
                   }
                     
                     for (int sx = 0;sx < shadSamps;sx++){
                         shadtar.y = 0;
                                    for (int sy = 0; sy < shadSamps; sy++){
                                    jitter = ((rand()%12501)/1000);
                                    posneg = rand()%2;
                                    jitter2 = ((rand()%12501)/1000);
                                    posneg2 = rand()%2;
                                    if(posneg==1){
                                                  jitter= -jitter;
                                                  }
                                                  if(posneg2==1){
                                                  jitter2= -jitter2;
                                                  }
                                    jittar = Vector3d(shadtar.x+jitter, shadtar.y+jitter2, 10);
                                    tempshade = GetShadeS(jittar, s, p);
                                    shadAVG[shadct]=tempshade;
                                    shadtar.y = shadtar.y - 6/shadSamps;
                                    shadct++;
                                    }
                         shadtar.x = shadtar.x + 6/shadSamps;
                         }
                         shadtar = Vector3d(6,3,10);
                         shadct=0;
                         for (int sx = 0;sx < shadSamps;sx++){
                         shadtar.y = 0;
                                    for (int sy = 0; sy < shadSamps; sy++){
                                    jitter = ((rand()%12501)/1000);
                                    posneg = rand()%2;
                                    jitter2 = ((rand()%12501)/1000);
                                    posneg2 = rand()%2;
                                    if(posneg==1){
                                                  jitter= -jitter;
                                                  }
                                                  if(posneg2==1){
                                                  jitter2= -jitter2;
                                                  }
                                    jittar = Vector3d(shadtar.x+jitter, shadtar.y+jitter2, 10);
                                    tempshade = GetShadeS(jittar, s2, p);
                                    shadAVG2[shadct]=tempshade;
                                    shadtar.y = shadtar.y - 6/shadSamps;
                                    shadct++;
                                    }
                         shadtar.x = shadtar.x + 6/shadSamps;
                         }
                 tempshade = 0;
                 tempshade2 = 0;
                 shade2 = 0;
                 for (int avg= 0; avg < ShadRays; avg++){
                     tempshade += shadAVG[avg];
                     tempshade2 += shadAVG2[avg];
                     }
                 shade = tempshade/ShadRays;
                 shade2 = tempshade2/ShadRays;
                 aacol[sa]=ColorRGB(aacol[sa].r*shade*shade2, aacol[sa].g*shade*shade2, aacol[sa].b*shade*shade2);
              }
       
               t2 = SphereIntersect(s2, r);
       if (t2 > 0&&t2 < t){
              t = t2;
              p = GetPP(t, r);
              sn = GetNormal(p, s2);
              ds = GetDS(alight, sn, p);
              if (ds > 0){
              aacol[sa] = ColorRGB(s2.mat.col.r*ds, s2.mat.col.g*ds, s2.mat.col.b*ds);
              }
              else{
                   aacol[sa] = ColorRGB(0,0,0);
                   }
              }
      t2 = TriIntersect(tri1, r);
      if (t2 > 0){
             t = t2;
             p = GetPP(t, r);
             aacol[sa] = tri1.mat.col;
             }
      t2 = TriIntersect(tri2, r);
      if (t2 > 0){
             t = t2;
             p = GetPP(t, r);
             aacol[sa]=tri1.mat.col;
             }
   r.d = backup;         
}
    pixcol.r = (aacol[0].r+aacol[1].r+aacol[3].r+aacol[2].r+aacol[4].r)/5;
    pixcol.g = (aacol[0].g+aacol[1].g+aacol[3].g+aacol[2].g+aacol[4].g)/5;
    pixcol.b = (aacol[0].b+aacol[1].b+aacol[3].b+aacol[2].b+aacol[4].b)/5;
    pset (x, y, fix(pixcol));
      scr.y += 0.01;   
    }
scr.z -=0.01;
   redraw();
}
    redraw();
    sleep();
}       
